/*
 * tokenize.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>

#include "config.h"

#include "hptsb.h"
#include "hptokdecl.h"
#include "decl.h"

#define TOKERROR(x,z) \
	{	ps->error_pos = lp; \
		ps->error_text = (uint8_t *) z; \
		return (x); \
	}

#define STORE(x) \
	{ \
		stlen += 1; \
		*stmt++ = ITOWH (x); \
		*stmt++ = ITOWL (x); \
	}

/*uint8_t * get_nth_character (uint8_t *, uint32_t);*/

uint32_t
tokenize_line (uint8_t * line, struct Program_Space *ps, struct Program_Entry *TPentry)
{
	int32_t		stlen;
	uint8_t		*lp = line;
	uint8_t		*nlp;
	uint8_t		*stmt;
	uint8_t		*nstmt;
	uint8_t		varop15 = 0;
	uint8_t		varname;
	uint8_t		vartype;
	uint8_t		save_varname;
	uint8_t		save_vartype;
	uint32_t	stringlen;
	uint8_t		fixstringlen;
	uint8_t		litstring[512];
	uint32_t	dimvar, dim1 = 0, dim2 = 0;
	uint32_t	varlast;
	uint32_t	lineno = 0;
	uint32_t	wlineno;
	uint16_t	ilineno;
	uint32_t	sttype;
	uint32_t	save_sttype;
	uint32_t	new_sttype;
	uint32_t	print_last_quoted;
	uint32_t	fileref;
	uint8_t		*statement;

	float		wnumber;
	uint32_t	HPfloat;

	statement = ps->statement;

	FNparse = 0;

	/* Determine sequence number */

	Dprintf (DEBUG_TOKEN, "Tokenizing line: %s\n", lp);
	if ((nlp = get_number (lp, GN_LINENO, &lineno)) == NULL)
		TOKERROR (1, "Statement number missing or invalid")

	lp = nlp;

	lp = match_token (lp, &sttype, Bst_ACCESS, Bst_ACCESS_count, 0x8000);
	if (sttype != 044 && sttype != 051)
		/* Only skip blanks after the statement type if not REM or IMAGE */
		lp = get_next_character (lp);

	if (sttype == 9999)	/* Assume "implied LET" if no operator match */
		sttype = 073;

	Dprintf (DEBUG_TOKEN, "*** Statement Type is: (%o) %s\n",
			sttype, Bst_ACCESS[sttype].ascii);

	/*
	 * Statement Syntax Parsing.
	 */

	stmt = ps->statement;
	stlen = 0;
	*stmt++ = lineno >> 8;	/* Statement No, upper byte */
	*stmt++ = lineno & 0xff;	/* Statement No, lower byte */
	stlen++;
	/* statement[2] and statement[3] (stlen) are set at the very end of parsing (statement len) */
	stmt++;		/* statement[2] */
	stmt++;		/* statement[3] */
	stlen++;

	switch (sttype) {
#ifdef TSC
	case 031:	/*******************************************/
			/* LIB - Special case for TSC TSB          */
			/*******************************************/
		lp = get_next_character (lp);
		break;
#endif
	case 032:	/*******************************************/
			/* SYSTEM                                  */
			/*******************************************/
		lp = get_next_character (lp);
		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
		}
		lp = nlp;

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript.")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (50, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		if (stringlen != 9999) {
			STORE (sttype << 9);
			sttype = 1;
			STORE (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);

			stmt += stringlen;
			stlen += stringlen / 2;
			if (stringlen % 2) {
				*stmt++ = 0;
				stlen++;
			}
		} else if (lp != nlp) {
			lp = nlp;

			nstmt = get_subscript (&nlp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			lp = nlp;
			stmt = nstmt;

		} else
			TOKERROR (49, "String expression expected")
		lp = nlp;
		break;

	case 033:	/*******************************************/
			/* CONVERT                                 */
			/*******************************************/

		nlp = get_string_variable (lp, &varname, &vartype);
		save_sttype = 0;
		if (lp != nlp) {
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else if ((nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen)) != lp && stringlen != 9999) {
			lp = nlp;
			if (sttype)
				STORE (sttype << 9);
			sttype = 1;
			fixstringlen = stringlen;

			STORE ((sttype << 9) | (fixstringlen & 0377));

			memcpy (stmt, litstring, stringlen);
			stlen = stlen + (stringlen + 1) / 2;
			stmt += stringlen;
			if (stringlen % 2)
				*stmt++ = 0;

			sttype = 0;
		} else if ((nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype)) != stmt) {
			save_sttype = 1;
			stmt = nstmt;
		} else
			TOKERROR (74, "Bad or missing CONVERT source")

		if ((nlp = match_literal (lp, "TO")) == NULL)
			TOKERROR (44, "TO expected")
		lp = nlp;

		sttype = 076;
				
		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
			else if (save_sttype)
				TOKERROR (46, "String variable expected")
		} else if (!save_sttype)
			TOKERROR (45, "Numeric variable expected");
		lp = nlp;

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		if (*lp != ',')
			break;
		sttype = 2;
		lp = get_next_character (lp + 1);

		/* Demand a statement number now */
		if ((lp = get_number (lp, GN_LINENO, &wlineno)) == NULL)
			TOKERROR (23, "Statement number missing or invalid")
		STORE (0x8000 | (sttype << 9) | 3);
		ilineno = wlineno;
		STORE (ilineno);

		break;

	case 034:	/*******************************************/
			/* LOCK                                    */
			/*******************************************/
	case 035:	/*******************************************/
			/* UNLOCK                                  */
			/*******************************************/
		STORE (sttype << 9);
		lp = get_next_character (lp);
		if (*lp != '#')
			TOKERROR (21, "'#' expected")
		lp = get_next_character(lp + 1);

		sttype = 4;
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (30, "File reference expected")
		stmt = nstmt;
		

		lp = get_next_character (lp);
		if (*lp != ',')
			break;
		lp = get_next_character (lp + 1);

		lp = get_simple_variable (lp, &varname, &vartype);
		if (!varname)
			TOKERROR (31, "Variable expected")
		sttype = 2;
		
		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;
		break;
	case 036:	/*******************************************/
			/* CREATE                                  */
			/*******************************************/
	case 037:	/*******************************************/
			/* PURGE                                   */
			/*******************************************/
		save_sttype = sttype;
		lp = get_simple_variable (lp, &varname, &vartype);
		if (!varname)
			TOKERROR (51, "Variable expected")

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (31, "Variable expected")
		lp = get_next_character (lp + 1);

		sttype = 2;
		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		lp = nlp;
		if (stringlen == 9999) {
			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else {
			STORE (sttype << 9);
			sttype = 1;
			STORE (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);

			stmt += stringlen;
			stlen += stringlen / 2;
			if (stringlen % 2) {
				*stmt++ = 0;
				stlen++;
			}
		}

		if (save_sttype == 037)		/* If PURGE, no more processing. */
			break;
		
		if (*lp != ',')
			TOKERROR (31, "Comma expected")
		lp = get_next_character (lp + 1);

		sttype = 2;
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (30, "Record length expression expected")
		stmt = nstmt;

		if (*lp != ',')
			break;
		lp = get_next_character (lp + 1);

		sttype = 2;
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (30, "Record length expected")
		stmt = nstmt;

		break;

	case 040: 	/*******************************************/
			/* ADVANCE                                 */
			/*******************************************/
		lp = get_next_character (lp);
		STORE (sttype << 9);
		lp = get_next_character (lp);
		if (*lp != '#')
			TOKERROR (21, "File reference expected")
		lp = get_next_character(lp + 1);

		sttype = 4;
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (34, "Bad or missing file reference")
		stmt = nstmt;
		lp = get_next_character (lp);


		if (*lp != ';')
			TOKERROR (33, "Semi-colon expected.")
		lp = get_next_character (lp + 1);

		sttype = 3;
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (34, "Missing or bad numeric value")
		stmt = nstmt;
		lp = get_next_character(lp);
		
		if (*lp != ',')
			TOKERROR (39, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;
		lp = get_simple_variable (lp, &varname, &vartype);
		if (!varname)
			TOKERROR (31, "Variable expected")

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		break;
	case 041:	/*******************************************/
			/* UPDATE                                  */
			/*******************************************/
		STORE (sttype << 9);
		lp = get_next_character (lp);
		if (*lp != '#')
			TOKERROR (21, "'#' File reference expected")
		lp = get_next_character(lp + 1);

		sttype = 4;
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (30, "File reference expected")
		stmt = nstmt;

		lp = get_next_character (lp);
		if (*lp != ';')
			TOKERROR (53, "Semi-colon expected")
		lp = get_next_character (lp + 1);
		sttype = 3;

		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		if (stringlen != 9999) {
			lp = nlp;
			STORE (sttype << 9);
			sttype = 1;
			STORE (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);

			stmt += stringlen;
			stlen += stringlen / 2;
			if (stringlen % 2) {
				*stmt++ = 0;
				stlen++;
			}
		} else if (lp != nlp) {
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else {

			stmt = nstmt;
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (30, "String or numeric expression expected")
			stmt = nstmt;
		}


		break;
	case 042:	/*******************************************/
			/* ASSIGN				   */
			/*******************************************/
		lp = get_next_character (lp);
		if (*lp == '*') {
			STORE (sttype << 9);

			sttype = 022;
			STORE (sttype << 9);
			lp = get_next_character (lp + 1);

			if (*lp != ',')
				TOKERROR (53, "Comma expected")
			lp = get_next_character (lp + 1);
			sttype = 2;

			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (30, "File reference expected")
			stmt = nstmt;

			if (*lp != ',')
				break;
			sttype = 2;
			lp = get_next_character (lp + 1);

			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Numeric variable expected")
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else {
			nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
			if (stringlen != 9999) {
				STORE (sttype << 9);
				sttype = 1;
				STORE (sttype << 9 | (stringlen & 0377));
				memcpy (stmt, litstring, stringlen);

				stmt += stringlen;
				stlen += stringlen / 2;
				if (stringlen % 2) {
					*stmt++ = 0;
					stlen++;
				}

				lp = nlp;
			} else if (lp != nlp) {
				lp = nlp;

				nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			} else
				TOKERROR (52, "String literal or variable expected")

			if (*lp != ',')
				TOKERROR (53, "Comma expected")
			lp = get_next_character (lp + 1);
			sttype = 2;

			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (30, "Missing file referencexx")
			stmt = nstmt;

			if (*lp != ',')
				TOKERROR (53, "Comma expectedxx")
			lp = get_next_character (lp + 1);
			sttype = 2;

			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			/* COMMA -- FILE NUMBER -- RETURN VARIABLE -- [ COMMA PROTECTION MASK ] -- */
			/*	[ COMMA RESTRICTION] */

			if (*lp != ',') {
				break;
			}
			lp = get_next_character (lp + 1);
			sttype = 2;

			nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
			if (nlp != lp) {
				if (stringlen != 9999) {
					lp = nlp;
					STORE (sttype << 9);
					sttype = 1;
					STORE (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);

					stmt += stringlen;
					stlen += stringlen / 2;
					if (stringlen % 2) {
						*stmt++ = 0;
						stlen++;
					}
				} else if (stringlen == 9999 && nstmt != NULL) {
					lp = nlp;
					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (30, "Bad or missing protection mask")
					stmt = nstmt;
				}
				lp = get_next_character (lp);
				if (*lp != ',')
					break;
				lp = get_next_character (lp + 1);
			}


			sttype = 2;
			STORE (sttype << 9);

			if ((nlp = match_literal (lp, "RR")) != NULL)
				sttype = 044;
			else if ((nlp = match_literal (lp, "WR")) != NULL)
				sttype = 045;
			else if ((nlp = match_literal (lp, "NR")) != NULL)
				sttype = 046;
			else
				TOKERROR (31, "Bad or missing restriction code")
			STORE (sttype << 9);
			lp = get_next_character (nlp);
			sttype = 0;
		}
		break;

	case 043:	/*******************************************/
			/* LINPUT                                  */
			/*******************************************/
		lp = get_next_character (lp);
		if (*lp == '#') {
			lp = get_next_character (lp + 1);

			STORE (sttype << 9);

			sttype = 4;

			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (30, "File reference expected")
			stmt = nstmt;

			if (*lp != ';')
				TOKERROR (55, "Semi-colon expected")
			lp = get_next_character (lp + 1);
			sttype = 3;
		}

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
		}
		lp = nlp;

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		break;

	case 044:	/*******************************************/
			/* IMAGE                                 */
			/*******************************************/
		if (*(lp - 1) == ' ')
			lp--;		/* Backup one character for IMAGE, I.E. "10 IMAGEABCD" is the formatting */
		STORE ((sttype << 9) | strlen((char *) lp));
		stlen += (strlen((char *) lp) / 2) + (strlen((char *) lp) % 2);
		memcpy(stmt, lp, strlen((char *) lp));		/* String */
		stmt += strlen((char *) lp);
		if (strlen((char *) lp) % 2)			/* If string spills over into */
			*(stmt++) = 0;			/*   next word, copy last byte */
		lp += strlen((char *) lp);
		break;
	case 045:	/*******************************************/
			/* COM                                     */
			/*******************************************/
	case 047:	/*******************************************/
			/* DIM                                     */
			/*******************************************/
		save_sttype = sttype;
		lp = get_next_character (lp);
		while (*lp) {
			dim1 = 0;
			dim2 = 0;
			dimvar = 0;
			nlp = get_string_variable (lp, &varname, &vartype);
			if (lp == nlp) {
				if (isalpha (*lp) && (*get_nth_character (lp, 1) == '(' || *get_nth_character (lp,1) == '[')) {
					dimvar = *lp;
					lp = get_next_character (lp + 1);

					if (*lp != '(' && *lp != '[')
						TOKERROR (61, "Missing left parenthesis")
					lp = get_next_character (lp + 1);

					nlp = get_number (lp, GN_INT, &dim1);
					if (nlp == NULL)
						TOKERROR (62, "Dimension expected")
					if (dim1 < 1 || dim1 > 9999)
						TOKERROR (80, "Illegal subscript")
					lp = get_next_character(nlp);

					if (*lp == ',') {
						lp = get_next_character (lp + 1);
						nlp = get_number (lp, GN_INT, &dim2);
						if (nlp == NULL)
							TOKERROR (64, "Invalid second dimension")
						if (dim2 < 1 || dim2 > 9999)
							TOKERROR (65, "Dimension out of range")
						lp = nlp;
					}

					if (*lp != ')' && *lp != ']')
					TOKERROR (66, "Missing dimension right parenthesis")
					lp = get_next_character (lp + 1);

					dimvar = dimvar - '@';
					STORE ((sttype << 9) | ((dimvar & 037) << 4) | ((dim2>0)?2:1));
					sttype = 012;	/* [ */
					STORE (0x8000 | (sttype << 9) | 3);
					STORE (dim1 & 0xFFFF);
					if (dim2 > 0) {
						sttype = 016;	/* subscript ',' comma */
						STORE (0x8000 | (sttype << 9) | 3);
						STORE (dim2 & 0xFFFF);
					}
					sttype = 011;	/* ] */
					STORE (sttype << 9);
				} else if (save_sttype == 045) {  /* Allow regular variables for COM */
					nlp = get_simple_variable (lp, &varname, &vartype);
					if (!varname)
						TOKERROR (73, "Invalid variable")
					lp = nlp;
					STORE ((sttype << 9) | ((varname & 037) << 4) | vartype);
				} else
					TOKERROR (60, "Invalid subscripted variable")
			} else {
				lp = get_next_character (nlp);

				if (*lp != '(' && *lp != '[')
					TOKERROR (67, "Left parenthesis expected")
				lp = get_next_character (lp + 1);

				nlp = get_number (lp, GN_INT, &dim1);
				if ( nlp == NULL || dim1 < 1 || dim1 > 9999)
					TOKERROR (68, "Illegal string dimension")
				lp = get_next_character(nlp);

				if (*lp != ')' && *lp != ']')
					TOKERROR (70, "Right parenthesis expected")
				lp = get_next_character (lp + 1);
			
				STORE ((sttype << 9) | ((varname & 037) << 4) | vartype);
				sttype = 012;   /* [ */
				STORE (0x8000 | (sttype << 9) | 3);
				STORE (dim1 & 0xFFFF);
				sttype = 011;   /* ] */
				STORE (sttype << 9);
			}
			if (*lp && *lp != ',')
				TOKERROR (71, "Comma expected")
			if (*lp == ',') {
				lp = get_next_character (lp + 1);
				if (!*lp)
					TOKERROR (72, "Extraneous comma")
				sttype = 002;
			}
		}
		break;

	case 050:	/*******************************************/
			/* DEF                                     */
			/*******************************************/
		lp = get_next_character (lp);
		/* Check for FNx */
		if (*lp != 'F' || *(get_next_character(lp + 1)) != 'N')
			TOKERROR (4, "FNx function name expected")
		vartype=017;
		lp = get_next_character (lp + 1);
		lp = get_next_character (lp + 1);

		if (!isalpha(*lp))
			TOKERROR (4, "Function name not FNA through FNZ")
		varname = toupper(*lp) - '@';
		STORE ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		lp = get_next_character (lp + 1);

		if (*lp != '(' && *lp != '[')
			TOKERROR (7, "Left parenthensis expected")
		sttype = 013;	/* Left Paren */
		lp = get_next_character (lp + 1);

		vartype = 4;
		if (!isalpha(*lp))
			TOKERROR (5, "Invalid user-defined function name")
		varname = toupper(*lp) - '@';
		lp = get_next_character (lp + 1);


		if (isdigit(*lp)) {
			vartype = 5 + (*lp - '0');
			lp = get_next_character (lp + 1);
		}

		FNparse = 1;
		FNvarname = varname;
		FNvartype = vartype;

		STORE (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		if (*lp != ')' && *lp != ']')
			TOKERROR (7, "Right parenthesis expected")

		sttype = 010;	/* Right Paren */
		varname = 0;
		vartype = 0;
		STORE ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		lp = get_next_character (lp + 1);
		if (*lp != '=')
			TOKERROR (6, "Assignment operator expected")
		lp = get_next_character (lp + 1);

		sttype = 017;	/* '=' assignment */
		varname = 0;
		vartype = 0;
		
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (35, "Invalid function defininition")
		stmt = nstmt;

		/* Every DEF FNx statement terminates with a zero-word */
		STORE (0);

		break;

	case 051:	/*******************************************/
			/* REM                                     */
			/*******************************************/
	case 070:	/*******************************************/
			/* FILES                                   */
			/*******************************************/

		STORE ((sttype << 9) | *lp);

		if (!*lp)
			break;
		lp++;

		/*stlen += (strlen((char *) lp) + 1) / 2;*/

		memcpy(stmt, lp, strlen((char *) lp));		/* String */
		stmt += strlen((char *) lp);
		stlen += (strlen((char *) lp) / 2);
		if (strlen((char *) lp) % 2) {
			*stmt++ = 0;
			stlen++;
		}

/*
		if ((strlen((char *) lp) % 2))
			*stmt++ = 0;
*/

		lp = lp + strlen((char *) lp);
		break;

	case 052:	/*******************************************/
			/* GOTO                                    */
			/*******************************************/
	case 056:	/*******************************************/
			/* GOSUB                                   */
			/*******************************************/
		lp = get_next_character (lp);
		nlp = get_number (lp, GN_LINENO, &wlineno);
		if (nlp != NULL && *(get_next_character(nlp)) == 0) {
			lp = nlp;
			/* We got a simple GOTO lineno */
			STORE (0x8000 | (sttype << 9) | 3);
			ilineno = wlineno;
			STORE (ilineno);
			sttype = 0;
			varname = 0;
			vartype = 0;
		} else {
			varname = 0;
			vartype = 0;
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (19, "Invalid GOTO syntax")
			stmt = nstmt;
			if (*(get_next_character(lp)) != 'O' || 
					*(get_nth_character(lp, 1)) != 'F')
				TOKERROR (20, "Invalid GOTO ... OF syntax")
			lp = get_next_character (lp + 1);
			lp = get_next_character (lp + 1);
			
			sttype = 074;  /* OF */
			STORE (0x8000 | (sttype << 9) | 3);

			lp = get_next_character(lp);
			while ((lp = get_number (lp, GN_LINENO, &wlineno)) != NULL) {
				ilineno = wlineno;
				STORE (ilineno);
				if (*lp == 0)
					break;
				if (*lp != ',')
					TOKERROR (21, "Comma expected")
				lp = get_next_character(lp + 1);
			}
			
		}

		break;

	case 053:	/*******************************************/
			/* IF                                      */
			/*******************************************/
		lp = get_next_character (lp);
		vartype = 0;
		varname = 0;

		if ((nlp = match_literal (lp, "END")) != NULL) {
			STORE (sttype << 9);
			/* END keyword found */
			lp = nlp;
			if (*lp != '#')
				TOKERROR (21, "'#' file reference expected")
			lp = get_next_character(lp + 1);
			sttype = 060;
			STORE (sttype << 9);
			sttype = 4;
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (37, "Invalid file reference")
			stmt = nstmt;
		} else if ((nlp = match_literal (lp, "ERROR")) != NULL) {
			STORE (sttype << 9);
			/* ERROR keyword found */
			sttype = 047;
			STORE (sttype << 9);
			lp = nlp;
		} else if ((nlp = get_string_variable (lp, &varname, &vartype)) != lp) {
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			sttype = 0;
			varname = 0;
			vartype = 0;

			nlp = get_relational_operator (lp, &sttype);
			if (lp == nlp)
				TOKERROR (57, "Relational operator expected")
			lp = nlp;

			if ((nlp = match_literal (lp, "CHR$")) != NULL) {
				lp = nlp;

				varname = 027;
				vartype = 017;
				STORE (0x8000 | (sttype << 9) | ((varname & 037)<<4) | (vartype & 017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
					TOKERROR (34, "Invalid or missing numeric argument to CHR$");
				stmt = nstmt;

				/* And a blank second argument for the TSB bug */
				STORE (0);

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				STORE (sttype << 9);
				sttype = 0;
			} else {

				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				Dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d *lp=%c\n",
						lp, nlp, varname, vartype, litstring, stringlen, *lp);
				lp = nlp;
				if (stringlen == 9999) {
					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
					Dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable, *lp=%c.\n", *lp);
				} else {
					STORE (sttype << 9);
					sttype = 1;
					STORE (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);

					stmt += stringlen;
					stlen += stringlen / 2;
					if (stringlen % 2) {
						*stmt++ = 0;
						stlen++;
					}
					Dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
				}
			}

		} else {
			Dprintf (DEBUG_TOKEN, "IF: Calling ans_formula: lp=%s\n", lp);
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (38, "Numeric expression expected")
			stmt = nstmt;
		}
			
		/* Demand THEN */
		if ((nlp = match_literal (lp, "THEN")) == NULL)
			TOKERROR (22, "THEN expected")
		sttype = 075;
		lp = get_next_character (nlp);

		/* Demand a statement number now */
		if ((lp = get_number (lp, GN_LINENO, &wlineno)) == NULL)
			TOKERROR (23, "Statement number missing or invalid")
		STORE (0x8000 | (sttype << 9) | 3);
		ilineno = wlineno;
		STORE (ilineno);
		break;

	case 054:	/*******************************************/
			/* FOR                                     */
			/*******************************************/
		lp = get_next_character (lp);
		/* Demand a simple variable */
		nlp = get_simple_variable (lp, &varname, &vartype);
		if (lp == nlp)
			TOKERROR (42, "Numeric variable expected")
		lp = get_next_character (nlp);
		STORE ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		varname = 0;
		vartype = 0;

		/* Demand an assignment '=' */
		if (*lp != '=')
			TOKERROR (56, "'=' expected")
		lp = get_next_character (lp + 1);
		sttype = 017;

		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (38, "Numeric expression expected")
		stmt = nstmt;

		/* Demand TO */
		if ((nlp = match_literal (lp, "TO")) == NULL)
			TOKERROR (57, "TO expected")
		lp = get_next_character(nlp);
		sttype = 076;

		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (38, "Numeric expression expected")
		stmt = nstmt;

		/* Optional STEP */
		lp = get_next_character (lp);
		if ((nlp = match_literal (lp, "STEP")) == NULL)
			break;
		lp = nlp;
		sttype = 077;

		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (38, "Numeric expression expected")
		stmt = nstmt;

		break;

	case 055:	/*******************************************/
			/* NEXT                                    */
			/*******************************************/
		lp = get_next_character (lp);
		lp = get_simple_variable (lp, &varname, &vartype);
		if (varname == 0)
			TOKERROR (27, "Numeric variable expected")
		STORE ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		break;
	case 057:	/*******************************************/
			/* RETURN                                  */
			/*******************************************/
	case 060:	/*******************************************/
			/* END                                     */
			/*******************************************/
	case 061:	/*******************************************/
			/* STOP                                    */
			/*******************************************/
		STORE (sttype << 9);
		break;
	case 062:	/*******************************************/
			/* DATA                                    */
			/*******************************************/
		lp = get_next_character (lp);
		while (*lp) {
			nlp = get_string_literal (lp, litstring, &stringlen);
			if (lp == nlp) {
				nlp = get_number (lp, GN_FLOAT, &wnumber);
				if (nlp == NULL)
					TOKERROR (58, "Invalid numeric or string expression for DATA")
				lp = nlp;
				Dprintf (DEBUG_TOKEN, "DATA: Storing a number: %G\n", wnumber);
				HPfloat = IEEEtoHPf(wnumber);
				Dprintf (DEBUG_TOKEN, "DATA: Checking wnumber=%G conversion: %G\n", wnumber, HPftoIEEE(HPfloat));
				STORE (0x8000 | (sttype << 9));
				STORE (HPfloat >> 16);
				STORE (HPfloat & 0xFFFF);
			} else {
				lp = nlp;
				STORE (sttype << 9);
				sttype = 1;
				STORE ((sttype << 9) | (stringlen & 0377));
				memcpy (stmt, litstring, stringlen);

				stmt += stringlen;
				stlen += stringlen / 2;
				if (stringlen % 2) {
					*stmt++ = 0;
					stlen++;
				}
			}
			sttype = 0;
			lp = get_next_character (lp);
			if (*lp != ',')
				break;
			sttype = 2;
			lp = get_next_character (lp + 1);
		}
		
		break;
	case 063:	/*******************************************/
			/* INPUT                                   */
			/*******************************************/
		lp = get_next_character (lp);
		while (*lp) {
			nlp = get_string_variable (lp, &varname, &vartype);
			if (lp != nlp) {
				lp = nlp;
			} else {
				nlp = get_simple_variable (lp, &varname, &vartype);
				if (lp == nlp)
					TOKERROR (34, "Variable expected")
				lp = nlp;
			}

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			if (*lp != ',')
				break;
			lp = get_next_character (lp + 1);
			sttype = 002;
		}
		break;
	case 064:	/*******************************************/
			/* READ                                    */
			/*******************************************/
		lp = get_next_character (lp);
		fileref = 0;
		if (*lp == '#') {
			STORE (sttype << 9);
			lp = get_next_character (lp + 1);

			sttype = 4;
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (34, "File reference expected")
			stmt = nstmt;

			if (*lp) {
				if (*lp == ',') {
					lp = get_next_character (lp + 1);
					sttype = 2;
					nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
					if (stmt == nstmt)
						TOKERROR (34, "Record number expected")
					stmt = nstmt;
				}
				if (*lp != ';') {
					if (!*lp)
						break;
					else
						TOKERROR (74, "Semi-colon expected")
				} else {
					sttype = 3;
					lp = get_next_character (lp + 1);
				}
				fileref = 1;
			} else
				break;		/* Allow simple "READ #n" statements. */
		}

		while (*lp) {
			nlp = get_string_variable (lp, &varname, &vartype);
			if (lp == nlp) {
				nlp = get_simple_variable (lp, &varname, &vartype);
				if (lp == nlp)
					TOKERROR (42, "Variable expected")
			}
			lp = nlp;
			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			if (*lp) {
				if (*lp != ',' && *lp != ';')
					TOKERROR (50, "List delimiter expected")
				lp = get_next_character (lp + 1);
			} else
				break;
			sttype = 002;
		}
		break;
	case 065:	/*******************************************/
			/* PRINT                                   */
			/*******************************************/
		lp = get_next_character (lp);
		print_last_quoted = 0;
		varlast = 0;
		fileref = 0;

		save_sttype = sttype;

		if (*lp == '#' ) {
			fileref = 1;
			STORE (sttype << 9);

			lp = get_next_character (lp + 1);

			sttype = 4;
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (34, "File reference expected")
			stmt = nstmt;

			if (*lp && *lp == ',') {
				lp = get_next_character (lp + 1);
				sttype = 2;
				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
					TOKERROR (34, "Record number expression expected")
				stmt = nstmt;
			}

			if (*lp) {
				if (*lp != ';')
					TOKERROR (91, "Semi-colon expected")
				sttype = 3;
				lp = get_next_character (lp + 1);
			}
		}

		if ((nlp = match_literal (lp, "USING")) != NULL) {
			STORE (sttype << 9);
			sttype = 043;

			lp = nlp;

			/* TODO: Get a FORMAT-STRING or a STATEMENT NUMBER */
			wlineno = 0;
			nlp = get_number (lp, GN_LINENO, &wlineno);
			if (nlp != NULL) {
				/* Statement number */
				STORE (0x8000 | (sttype << 9) | 3);
				lp = nlp;
				ilineno = wlineno;
				STORE (ilineno);
			} else {
				/* TODO: Get a FORMAT-STRING */
				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				if (stringlen == 9999) {
					lp = nlp;

					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;

				} else {
					lp = nlp;
					STORE (sttype << 9);
					sttype = 1;
					STORE (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);
					stlen += (stringlen + 1) / 2;
					stmt += stringlen;
					if (stringlen % 2) {
						*stmt++ = 0;
					}
				}
				sttype = 0;
			}

			if (*lp) {
				if (*lp != ';')
					TOKERROR (91, "Semi-colon expected")
				sttype = 3;
				lp = get_next_character (lp + 1);
			} else
				break;
		}

		nlp = lp;	/* Needed to save lp prior to get_ansformula below */
		while (*lp) {
			stringlen = 0;
			if ((nlp = match_literal (lp, "UPS$")) != NULL) {
				if (varlast)
					TOKERROR (13, "1-List delimiter expected");
				varlast = 1;
				lp = nlp;

				varname = 030;
				vartype = 017;
				STORE (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype&017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				if (stringlen != 9999) {
					STORE (sttype << 9);
					sttype = 1;
					STORE (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);

					stmt += stringlen;
					stlen += stringlen / 2;
					if (stringlen % 2) {
						*stmt++ = 0;
						stlen++;
					}
					lp = nlp;
				} else if (nlp != lp) {
					lp = nlp;

					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
				}

				/* And a blank second argument for the TSB bug */
				STORE (0);

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				STORE (sttype << 9);
				sttype = 0;

				print_last_quoted = 0;

			}  else if ((nlp = match_literal (lp, "CHR$")) != NULL) {
				if (varlast)
					TOKERROR (13, "2-List delimiter expected")
				varlast = 1;
				lp = nlp;

				varname = 027;
				vartype = 017;
				STORE (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype&017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
				TOKERROR (34, "Expecting numeric argument to CHR$")
				stmt = nstmt;

				/* And a blank second argument for the TSB bug */
				STORE (0);

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				STORE (sttype << 9);
				sttype = 0;

				print_last_quoted = 0;

#ifndef OLD_WORKINGFIXFORSPLITSTRINGLITERALS
			}  else if ((nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen)) != lp
					 			&& stringlen != 9999) {
#else
			} else if ((nlp = get_string_literal (lp, litstring, &stringlen)) != lp
								&& stringlen != 9999) {

				/* This check was for 'xx literals only, as TSB will separate them from "'s */
				/* stringlen was set at the beginning of the while(), and the call resets it */

				lp = nlp;
				if (sttype) {
					STORE (sttype << 9);
				}
				varlast = 0;

				sttype = 1;
				fixstringlen = stringlen;

				STORE ((sttype << 9) | (fixstringlen & 0377));

				memcpy (stmt, litstring, stringlen);
				stmt += stringlen;
				if (stringlen % 2) {
					*stmt++ = 0;
				}
				sttype = 0;

				print_last_quoted = 0;
			}  else if ((nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen)) != lp
					 			&& stringlen != 9999) {
#endif
				if (print_last_quoted)
					TOKERROR (13, "3-List delimiter expected")
				lp = nlp;
				if (sttype) {
					STORE (sttype << 9);
				}

				sttype = 1;
				fixstringlen = stringlen;

				STORE ((sttype << 9) | (fixstringlen & 0377));

				memcpy (stmt, litstring, stringlen);
				stlen = stlen + (stringlen + 1) / 2;
				stmt += stringlen;
				if (stringlen % 2) {
					*stmt++ = 0;
				}
				sttype = 0;

				print_last_quoted = 1;

			} else if (nlp != lp) {
				lp = nlp;

				nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;

				print_last_quoted = 0;

			} else if (save_sttype == 065 && (nlp = match_literal (lp, "END")) != NULL) {
				if (!fileref)
					TOKERROR (13, "END invalid without file reference");
				if (varlast)
					TOKERROR (13, "4-List delimiter expected");
				STORE (sttype << 9);
				sttype = 060;
				lp = nlp;
				break;
			} else {
#ifdef YUCK
fprintf (stderr, "*********** CHECKING FOR FUNCTION ***********\n"); fflush (stderr);
				nlp = match_token (lp, &new_sttype, 
                                        	MACH_TYPE_TSC(tape_mach_type) ? Bfunc_TSC : Bfunc,
                                        	MACH_TYPE_TSC(tape_mach_type) ? BfuncTSC_count : Bfunc_count, 0);

				if (nlp != lp && new_sttype <= 3) {
#ifdef OLD /* Removed 2012/09/15 -- Disallowed "..."TAB(65) without a delimiter */
					if (varlast)
						TOKERROR (13, "5-List delimiter expected");
#endif
					varlast = 1;
					/*sttype = new_sttype;*/
					STORE (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);
					sttype = 0;
					lp = nlp;

					if (*lp != '(' && *lp != '[')
						TOKERROR (93, "Left parenthesis expected")

					sttype = 013;
					lp = get_next_character (lp + 1);

					nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
					if (stmt == nstmt)
						TOKERROR (95, "Numeric expression expected")
					stmt = nstmt;

					/* And a blank word pair for the TSB funkiness */
					STORE (0);

					if (*lp != ')' && *lp != ']')
						TOKERROR (96, "Right parentheses expected")

					sttype = 010;
					STORE (sttype << 9);
					sttype = 0;
					lp = get_next_character (lp + 1);

					print_last_quoted = 0;
					
				} else
				    if ((nstmt = ans_formula (&nlp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype)) != stmt) {
#endif
                                /*sttype = 0;*/
                                varname = 0;
                                vartype = 0;
				nlp = lp;

				    if ((nstmt = ans_formula (&nlp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype)) != stmt) {
#ifdef OLD /* Removed 2012/09/15 -- Disallowed "..."N[1] without a delimiter */
					if (varlast)
						TOKERROR (13, "6-List delimiter expected");
#endif
					varlast = 1;
					lp = nlp;
					stmt = nstmt;
					sttype = 0;

					print_last_quoted = 0;

				} else {
					nlp = get_simple_variable (lp, &varname, &vartype);
					if (lp != nlp) {
						lp = nlp;

						nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
						if (nstmt == NULL)
							TOKERROR (80, "Invalid subscript")
						stmt = nstmt;

						print_last_quoted = 0;

					} else
						TOKERROR (98, "Invalid syntax")
				}
			}

			if (*lp == ';') {
				varlast = 0;
				lp = get_next_character (lp + 1);
				if (sttype) {
					STORE (sttype << 9);
				}
				sttype = 3;
			} else if (*lp == ',') {
				varlast = 0;
				lp = get_next_character (lp + 1);
				if (sttype) {
					STORE (sttype << 9);
				}
				sttype = 2;
			} else if (!*lp)
				break;
			else {
				lp = get_next_character (lp);
				if (*lp && print_last_quoted &&
						(*get_next_character (lp) == '('
						|| *get_next_character (lp) == '['
						|| *get_next_character (lp) == '+'
						|| *get_next_character (lp) == '-')) {
					STORE (0);
				}
				sttype = 0;	/* List delimiters are not mandated in PRINT after "..." */
						/* BUG: "..." can't follow "..." without a delimiter */
			}
			print_last_quoted = 0;
		}

		if (sttype) {
			STORE (sttype << 9);
		}

		break;
	case 066:	/*******************************************/
			/* RESTORE                                 */
			/*******************************************/
		lp = get_next_character (lp);
		wlineno = 0;
		nlp = get_number (lp, GN_LINENO, &wlineno);
		if (nlp == NULL) {
			if (*lp)
				TOKERROR (29, "Statement number missing or invalid")
			STORE (sttype << 9);
		} else {
			/* Optional statement number */
			STORE (0x8000 | (sttype << 9) | 3);
			lp = nlp;
			ilineno = wlineno;
			STORE (ilineno);
		}
		break;
	case 067:	/*******************************************/
			/* MAT                                     */
			/*******************************************/
		/* TODO */
		/*
		 * MAT PRINT				<array-variable> | <print-function> [ ,  ; ... ]
		 * MAT PRINT [ #<numeric-expression ;]	[ <array-variable> | <print-function> ] [ ,  ; ... ] [END]
		 * MAT PRINT USING  <format-expression>	<array-variable> | <print-function> [ ,  ; ... ]
		 * MAT READ [ #<numeric-expression> ;]	<array-variable> [ ( <dimensions ) ] [ ... ]
		 * MAT INPUT 				<array-variable> [ ( <dimensions ) ] [ ... ]
		 * MAT  				<array-variable> = { ZER | CON | IDN }
		 * MAT					<array-variable> = <array-variable> [ +|-|* <array-variable> ]
		 * MAT					<array-variable> = {TRN | INV} ( <array-variable> )
		 */
		
		if ((nlp = match_literal (lp, "PRINT")) != NULL) {
			STORE (sttype << 9);
			sttype = 065;
			lp = nlp;
			if ((nlp = match_literal (lp, "USING")) != NULL) {
				STORE (sttype << 9);
				sttype = 043;
				lp = nlp;

				/* TODO: Get a FORMAT-STRING or a STATEMENT NUMBER */
				wlineno = 0;
				nlp = get_number (lp, GN_LINENO, &wlineno);
				if (nlp != NULL) {
					/* Statement number */
					STORE (0x8000 | (sttype << 9) | 3);
					lp = nlp;
					ilineno = wlineno;
					STORE (ilineno);
				} else {
					/* TODO: Get a FORMAT-STRING */
					nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
					if (stringlen == 9999) {
						lp = nlp;
	
						nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
						if (nstmt == NULL)
							TOKERROR (80, "Invalid subscript")
						stmt = nstmt;
	
					} else {
						lp = nlp;
						STORE (sttype << 9);
						sttype = 1;
						STORE (sttype << 9 | (stringlen & 0377));
						memcpy (stmt, litstring, stringlen);
						stlen += (stringlen + 1) / 2;
						stmt += stringlen;
						if (stringlen % 2) {
							*stmt++ = 0;
						}
					}
					sttype = 0;
				}
				if (*lp) {
					if (*lp != ';')
						TOKERROR (91, "Semi-colon expected")
					sttype = 3;
					lp = get_next_character (lp + 1);
				} else
					break;
			}
		} else if ((nlp = match_literal (lp, "READ")) != NULL) {
			STORE (sttype << 9);
			sttype = 064;
			lp = nlp;
		} else if ((nlp = match_literal (lp, "INPUT")) != NULL) {
			STORE (sttype << 9);
			sttype = 063;
			lp = nlp;
		}

		save_sttype = sttype;	/* Keep track of what MAT format we're parsing. */
		new_sttype = 0;		/* Keep track of whether we're reading or writing a file. */

		if ((sttype == 065 || sttype == 064) && *lp == '#') {
			new_sttype = 1;
			STORE (sttype << 9);
			sttype= 4;
			lp = get_next_character (lp + 1);

			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (34, "Missing or invalid file reference")
			stmt = nstmt;

			if (*lp == ',') {
				lp = get_next_character (lp + 1);
				sttype = 2;

				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
					TOKERROR (34, "Missing or invalid record number")
				stmt = nstmt;
			}

			if (*lp != ';')
				TOKERROR (38, "Semi-colon expected")
			lp = get_next_character (lp + 1);
			if (sttype)
				TOKERROR (-99, "INTERNAL ERROR #MAT1: left-over sttype undealt with")
			sttype= 3;
		}
			
		if (save_sttype == 0 && new_sttype == 1) {
			/* allow END or a print-function for 'MAT PRINT #' */
			/* Just fall through for a print-function, but process and finish for END */
			/* otherwise error */
		}

		if (save_sttype == 065 && new_sttype == 1 && !*lp)
			break;

		nlp = get_simple_variable (lp, &varname, &vartype);
		if (lp != nlp) {

			lp = nlp;
			vartype = 3;
			if (*lp == '[' || *lp == '(') {
				stlen |= 0x4000;
			}

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else
			TOKERROR (37, "Array variable expected")


		if (save_sttype == 067) {	/* Plain old MAT =? */
			if (!*lp)
				break;
			if (*lp != '=')
				TOKERROR (36, "'=' assignment operator expected")
			sttype = 017;		/* Assignment '=' */
			lp = get_next_character (lp + 1);

			if ((nlp = match_token (lp, &new_sttype,
                                        MACH_TYPE_TSC(tape_mach_type) ? Bfunc_TSC : Bfunc,
                                        MACH_TYPE_TSC(tape_mach_type) ? BfuncTSC_count : Bfunc_count, 5)) != lp) {

				/* Got a MAT assignment function */

				if (new_sttype >= 033 && new_sttype <= 035) {

					lp = nlp;

					varname = new_sttype;
					vartype = 017;
					varop15 = 1;  /* Need to ensure 0x8000 is biased in get_subscript */
					stlen |= 0x4000; /* And set flag to handle function-type subscripts */

					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
					break;
				} else if (new_sttype < 033)
					TOKERROR (81, "Invalid function for MAT")

				lp = nlp;
				if (*lp != '[' && *lp != '(')
					TOKERROR (81, "Left parenthesis expected")

				lp = nlp;
				STORE (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);

				sttype = 0;
				varname = 0;
				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
					TOKERROR (82, "Numeric expression expected")
				stmt = nstmt;
				break;
			}
			if (*lp == '[' || *lp == '(') {
				STORE (sttype << 9);
				/* Handle "MAT A=(numeric formula) * B */
				sttype = 013;
				lp = get_next_character (lp + 1);

				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
						TOKERROR (83, "Numeric expression expected")
				stmt = nstmt;
				if (*lp != ']' && *lp != ')')
					TOKERROR (79, "Missing right parenthesis")
				lp = get_next_character (lp + 1);

				sttype = 010;
				STORE (sttype << 9);

				if (*lp != '*')
					TOKERROR (79, "'*' expected")
				lp = get_next_character (lp + 1);
				sttype = 022;

				nlp = get_simple_variable (lp, &varname, &vartype);
				if (lp != nlp) {
					lp = nlp;

					vartype = 3;
					stlen |= 0x4000;
					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
				} else
					TOKERROR (79, "Array variable expected")
				break;
			}
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
					TOKERROR (83, "Numeric expression expected")
			stmt = nstmt;
		}
		
		while (*lp) {
			if (!*lp)
				break;
			if (*lp == ',') {
				sttype = 2;
			} else if (*lp == ';') {
				sttype = 3;
			} else
				TOKERROR (50, "List delimiter expected")
			lp = get_next_character (lp + 1);
			if (!*lp)
				break;

			if (save_sttype == 065 && new_sttype == 1 && (nlp = match_literal (lp, "END")) != NULL) {
				STORE (sttype << 9);
				sttype = 060;
				lp = nlp;
				break;
			}
				
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp != nlp) {
				lp = nlp;

				vartype = 3;
				stlen |= 0x4000;
				nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			} else
				TOKERROR (84, "Invalid MAT PRINT source")
		}
		if (sttype)
			STORE (sttype << 9);
		
		break;

	case 071:	/*******************************************/
			/* CHAIN                                   */
			/*******************************************/

		/* TODO: [<return-variable> , ] <program-name> [, <numeric-expression> ] */

		/* Dummy call to get_source_string to see if a string follows */
		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);

		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp != nlp) {
				lp = nlp;

				nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;

				if (*lp != ',')
					TOKERROR (50, "Comma expected")
				lp = get_next_character (lp + 1);
				sttype = 2;
			}
		}

		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		if (stringlen != 9999) {
			STORE (sttype << 9)
			sttype = 1;
			STORE (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);

			stmt += stringlen;
			stlen += stringlen / 2;
			if (stringlen % 2) {
				*stmt++ = 0;
				stlen++;
			}
			lp = nlp;
		} else if (varname != 0) {
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

		} else
			TOKERROR (49, "String expression expected for program name")

		lp = get_next_character (lp);
		if (!*lp)
			break;
		if (*lp != ',')
			TOKERROR (50, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (51, "Numeric expression expected")
		stmt = nstmt;
		break;

	case 072:	/*******************************************/
			/* ENTER                                   */
			/*******************************************/

		lp = get_next_character (lp);

		if (*lp == '#') {
			lp = get_next_character (lp + 1);
			/* Store the ENTER statement code */
			STORE (sttype << 9);

			sttype = 4;
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (39, "Numeric variable expected")
			lp = nlp;

			nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
			if (*lp == ',')
				sttype = 2;
			else {
				if (!*lp)
					break;
				TOKERROR (39, "Comma expected")
			}
			lp = get_next_character (lp + 1);
		}
			
		nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
		if (stmt == nstmt)
			TOKERROR (34, "Bad or missing Time Allowed expression")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (39, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nlp = get_simple_variable (lp, &varname, &vartype);
		if (lp == nlp)
			TOKERROR (34, "Numeric variable expected")
		lp = nlp;

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (39, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected for returned value")
		}
		lp = nlp;

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		break;

	case 046:	/*******************************************/
			/* LET                                     */
			/*******************************************/
	case 073:	/*******************************************/
			/* Implied LET                             */
			/*******************************************/

		lp = get_next_character (lp);
		if (!isalpha(*lp))
			TOKERROR (3, "Unknown statement type, assignment operator expected");
		varname = 0;
		vartype = 0;

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
		}
		lp = nlp;
		if (vartype == 4 && (*lp == '[' || *lp == '('))
			vartype = 4;  /* Tweak for the vartype going into get_subscript */

		save_vartype = vartype;
		save_varname = varname;
		varop15 = 0;

		nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		lp = get_next_character (lp);
		if (*lp != '=')
			TOKERROR (-10, "Unknown statment, assignment operator expected")
		lp = get_next_character (lp + 1);
		sttype = 017;		/* Type is now '=' (assignment) */

		if (save_vartype == 0 || (save_varname >= 033 && save_varname <= 036)) {
		
			nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
			if (stringlen != 9999) {
				STORE (sttype << 9);
				sttype = 1;
				STORE (sttype << 9 | (stringlen & 0377));
				memcpy (stmt, litstring, stringlen);

				stmt += stringlen;
				stlen += stringlen / 2;
				if (stringlen % 2) {
					*stmt++ = 0;
					stlen++;
				}
				lp = nlp;
			} else if (lp != nlp) {
				lp = nlp;

				nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;

			} else if ((nlp = match_literal (lp, "UPS$")) != NULL) {
				lp = nlp;

				varname = 030;
				vartype = 017;
				STORE (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype&017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				if (stringlen != 9999) {
					STORE (sttype << 9);
					sttype = 1;
					STORE (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);

					stmt += stringlen;
					stlen += stringlen / 2;
					if (stringlen % 2) {
						*stmt++ = 0;
						stlen++;
					}
					lp = nlp;
				} else if (nlp != lp) {
					lp = nlp;

					nstmt = get_subscript (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype, &varop15);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
				}

				/* And a blank second argument for the TSB bug */
				STORE (0);

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				STORE (sttype << 9);

			}  else if ((nlp = match_literal (lp, "CHR$")) != NULL) {
				lp = nlp;

				varname = 027;
				vartype = 017;
				STORE (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype&017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
				if (stmt == nstmt)
					TOKERROR (34, "Invalid argument to CHR$")
				stmt = nstmt;

				/* And a blank second argument for the TSB bug */
				STORE (0);

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				STORE (sttype << 9);
				sttype = 0;

			} else
				TOKERROR (97, "Invalid source string")

		} else {
			sttype |= 0x8000;
			nstmt = ans_formula (&lp, &sttype, stmt, (uint32_t *) &stlen, &varname, &vartype);
			if (stmt == nstmt)
				TOKERROR (34, "Invalid assignment value")
			stmt = nstmt;
		}

		Dprintf (DEBUG_TOKEN, "NEW stlen=%d\n", stlen);
		break;
	}

	if (*lp)
		TOKERROR (24, "Characters after statement end")

	ps->statement[2] = ITOWH (stlen);			/* Statement len, upper byte */
	ps->statement[3] = ITOWL (stlen);			/* Statement len, lower byte */

	Dprintf (DEBUG_TOKEN, "Statement length is: %d words, %dbytes\n",
			WTOI (statement + 2) + 1, WTOI(statement + 2) * 2 + 2);

	ps->statement_len = stlen;

	return (0);
		
}

/*
 *	match_literal - Checks for characters 'literal' in 'line'.
 *			returns a pointer to the next non-blank
 *			character after a match, or NULL if unmatching.
 */

uint8_t *
match_literal (uint8_t *line, char *literal)
{
	uint32_t	i;
	uint8_t		*nlp;

	nlp = line;
	for (i = 0; i < strlen(literal); i++) {
		nlp = get_next_character(nlp);
		if (!*nlp || *nlp != literal[i])
			return NULL;
		nlp++;
	}
	return (get_next_character(nlp));
}

/*
 * match_token - Check the next characters for a match agains a
 *		 'struct BASIC_Tokens' table.  Sets opcode to
 *		 the operand code.  Btok points to the table base,
 *		 count indicates the number of table entries,
 *		 and prio indicates the minimum table "priority"
 *		 entry to be checked.
 */

uint8_t *
match_token (uint8_t * line, uint32_t *opcode, struct BASIC_Tokens *Btok, uint32_t count, uint32_t prio)
{
	int32_t		i, j;
	uint8_t		*nlp;
	uint32_t	toklen;
	uint8_t		*tokstart;
	uint32_t	noskipblanks = prio & 0x8000;

	if (noskipblanks)
		prio = prio & 0x7fff;
#ifdef SEVEREDEBUGGING
	Dprintf (DEBUG_TOKEN, "match_token: Looking up item in table, count=%d (%o)\n", count, count);
#endif
	for (i = count - 1; i >= 0; i--) {
		nlp = line;

#ifdef SEVEREDEBUGGING
		Dprintf (DEBUG_TOKEN, "match_token: Checking table item #%d (%o)\n", i, i);
#endif
		if (Btok[i].ascii[0] == 0)
			continue;
		if (Btok[i].priority < prio) {
			continue;
		}

#ifdef SEVEREDEBUGGING
		Dprintf (DEBUG_TOKEN, "match_token: Comparing %-.3s to %s...\n", line, Btok[i].ascii);
#endif

		tokstart = (uint8_t *) Btok[i].ascii;
		if (*tokstart == ' ')
			tokstart++;
		toklen = (uint32_t) strlen((char *) tokstart);
		if (*(tokstart + toklen - 1) == ' ')
			toklen--;

		for (j = 0; j < toklen; j++) {
			nlp = get_next_character (nlp);
			if (*nlp != *(tokstart + j)) {
				nlp = NULL;
				break;
			}
			nlp++;
		}
		if (nlp != NULL) {
			Dprintf (DEBUG_TOKEN, "match_token: MATCH FOUND, tokencode = %o: %s\n", i, Btok[i].ascii);
			*opcode = Btok[i].op;
			return (noskipblanks ? nlp : get_next_character(nlp));
		}
	}
	Dprintf (DEBUG_TOKEN, "match_token: NO MATCH FOUND for %-.3s\n", line);
	*opcode = 9999;
	return line;
}

/*
 * get_next_character - return the first non-blank character
 *			position.
 */

uint8_t *
get_next_character (uint8_t *line)
{
	while (*line != 0 && (*line == ' ' || *line == '\t'))
		line++;
	return line;
}

/*
 * get_nth_character - return the n-th non-blank character
 *			position.
 */

uint8_t *
get_nth_character (uint8_t *line, uint32_t charno)
{
	uint8_t		*lp = line;
	uint32_t	i;
	for (i = 0; i < charno; i++) {
		while (*lp != 0 && (*lp == ' ' || *lp == '\t'))
			i++;
		if (*lp == 0)
			break;
		lp++;
	}
	if (i == charno)
		return (lp);
	return (NULL);
}

/*
 * get_simple_variable - Expect and return a numeric variable in varname
 *			 and vartype.  Subscripts are not processed.
 */

uint8_t *
get_simple_variable (uint8_t *lp, uint8_t *varname, uint8_t *vartype)
{
	lp = get_next_character (lp);

	*varname = 0;
	*vartype = 0;

	if (!isalpha(*lp)) {
	Dprintf (DEBUG_TOKEN, "get_simple_variable:  non-alpha char, not bothering: '%c'\n", *lp);
		return (lp);
	}
	*vartype = 4;
	*varname = toupper(*lp) - '@';
		
	lp = get_next_character (lp + 1);

	if (isdigit(*lp)) {
		*vartype = 5 + (*lp - '0');
		lp = get_next_character (lp + 1);
	}

	Dprintf (DEBUG_TOKEN, "get_simple_variable: varname=%c vartype = %d\n", *varname + '@', *vartype);
	return (lp);
}

/*
 * get_string_variable - Expect and return the varname and vartype of
 *			 a string variable.  Subscripts are not processed.
 */

uint8_t *
get_string_variable (uint8_t *lp, uint8_t *varname, uint8_t *vartype)
{
	uint8_t		varnum;
	uint8_t		*nlp;
	nlp = get_next_character (lp);

	*varname = 0;
	*vartype = 0;

	if (!isalpha(*nlp))
		return (lp);
	*varname = *nlp;
	nlp = get_next_character (nlp + 1);
	if (isdigit (*nlp)) {
		varnum = *nlp - '0';
		*vartype = 5;
		if (varnum == 0) {
			if (*varname < 'Q') {
				*vartype = *varname - 'A';
				*varname = 033;
			} else {
				*vartype = *varname - 'Q';
				*varname = 034;
			}
		} else if (varnum == 1) {
			if (*varname < 'Q') {
				*vartype = *varname - 'A';
				*varname = 035;
			} else {
				*vartype = *varname - 'Q';
				*varname = 036;
			}
		}
#ifdef DEBUGASCII /* Removed, as this message was output for DIM P9, complaining about the 9, but tokenizing OK */
		 else 
			eprintf ("BOGUS varnum: %d: %s\n", varnum, lp);
#endif
		nlp = get_next_character (nlp + 1);
	} else {
		*varname = *varname - '@';
	}

	if (*nlp != '$') {
		*varname = 0;
		return (lp);
	}
	nlp = get_next_character (nlp + 1);
	Dprintf (DEBUG_TOKEN, "get_string_variable: Got variable, varname=%d  vartype=%o\n", *varname, *vartype);
	return (nlp);
}

/*
 * get_subscript - expect a subscript notation, and analyze and store
 *		   it when found.  sttype should be biased by 0x8000
 *		   if parenthesis should be stored instead of brackets
 *		   (for functions, etc).  If sttype is biased by 0x4000,
 *		   then this is a function definition LHS, and an
 *		   additional empty word should be stored after the
 *		   encapsulated parameter to match TSB oddness.
 */

uint8_t *
get_subscript (uint8_t **lp, uint32_t *sttype, uint8_t *stmt, uint32_t *stlen, uint8_t *varname, uint8_t *vartype, uint8_t *varop15)
{
	uint8_t		*nlp;
	uint32_t	save_stlen;
	uint8_t		*save_stmt;
	uint8_t		*save_lp;
	uint8_t		*nstmt;
	uint8_t		*VARstmt;
	uint8_t		VARop15, VARname, VARtype;
	uint32_t	VARsttype;
	uint32_t	parentype = 0;
	uint32_t	functype = 0;
	uint32_t	wrotefirst = 0;

	parentype = (*stlen & 0x8000); /* if biased by 0x8000, uses parenthesis instead of brackets */
	functype  = (*stlen & 0x4000);
	*stlen &= 0x3fff;

	/* Store, but remember the varname/vartype pair.   We'll have to change the vartype later. */

	VARsttype = *sttype;
	VARop15 = *varop15;
	VARname = *varname;
	VARtype = *vartype;
	VARstmt = stmt;

	if (FNparse && VARname == FNvarname && VARtype == FNvartype)
		VARop15 = 1;

	if (VARsttype || VARname || VARtype) {
		*stlen += 1;
		*VARstmt++ = ITOWH ((VARop15 << 15) | (VARsttype << 9) | ((VARname & 037) << 4) | (VARtype & 017));
		*VARstmt++ = ITOWL ((VARop15 << 15) | (VARsttype << 9) | ((VARname & 037) << 4) | (VARtype & 017));
		VARstmt = stmt;
		stmt++;
		stmt++;

		wrotefirst++;
	}

	save_stlen = *stlen;
	save_stmt = stmt;
	save_lp = *lp;
	
	*sttype = 0;
	*varname = 0;
	*vartype = 0;
	*varop15 = 0;

	Dprintf (DEBUG_TOKEN, " PRE-lp=%x\n", *lp);
	nlp = *lp;
	*lp = get_next_character (*lp);

	if (**lp != '(' && **lp != '[')
		return (stmt);
	if ((VARtype > 4 && (VARname < 033 || VARname > 036)) && VARtype != 017) {
			return (stmt);
	}

	*sttype = parentype ? 013 : 012;
	
	*lp = get_next_character (*lp + 1);

	nstmt = ans_formula (lp, sttype, stmt, stlen, varname, vartype);

	Dprintf (DEBUG_TOKEN, "get_subscript: YYY ans_formula returned, **lp=%c  stmt=%x  nstmt=%x\n", **lp, stmt, nstmt);

	if (stmt == nstmt) {
		*sttype = 0;
		*stlen = save_stlen;
		stmt = save_stmt;
		*lp = save_lp;
		*lp = nlp;
		return (stmt);
	}
	stmt = nstmt;
	if ( ((VARname < 033 || VARname >= 036) && VARtype == 4) || VARtype == 3)
		VARtype = 1;		/* 1-subscript variable */

	if (**lp == ',') {
		if (*sttype != 0) {
			*stlen += 1;
			*stmt++ = ITOWH ((*sttype) << 9);
			*stmt++ = ITOWL ((*sttype) << 9);
		}
		*lp = get_next_character (*lp + 1);
		*sttype = 016;

		nstmt = ans_formula (lp, sttype, stmt, stlen, varname, vartype);
		Dprintf (DEBUG_TOKEN, "get_subscript: ZZZ ans_formula returned, **lp=%c,  stmt=%x  nstmt=%x\n", **lp);

		if (stmt == nstmt) {
				
			*sttype = 0;
			*stlen = save_stlen;
			stmt = save_stmt;
			*lp = save_lp;
			*lp = nlp;
			return (stmt);
		}
		stmt = nstmt;
		if (VARtype == 1)	/* 2-subscript variable */
			VARtype = 2;
	}

	/* If a function, add an empty word */
	if (functype) {
		*stlen += 1;
		*stmt++ = 0;
		*stmt++ = 0;
	}

	if ((**lp != ')' && **lp != ']')) {
		*sttype = 0;
		*stlen = save_stlen;
		stmt = save_stmt;
		*lp = save_lp;
		return (stmt);
	}

	*sttype = parentype ? 010 : 011;
	*lp = get_next_character (*lp + 1);
		
	*stlen += 1;
	*stmt++ = ITOWH ((*sttype) << 9);
	*stmt++ = ITOWL ((*sttype) << 9);

	/* Update the vartype */
	if (wrotefirst) {
		*VARstmt++ = ITOWH ((VARop15 << 15) | (VARsttype << 9) | ((VARname & 037) << 4) | (VARtype & 017));
		*VARstmt++ = ITOWL ((VARop15 << 15) | (VARsttype << 9) | ((VARname & 037) << 4) | (VARtype & 017));
	}

	*sttype = 0;
	*varop15 = 0;
	*varname = 0;
	*vartype = 0;

	Dprintf (DEBUG_TOKEN, "get_subscript: PUT A SUBSCRIPT, next position: '%c' -- stlen=%d stmt=%x\n", **lp, *stlen, stmt);
	Dprintf (DEBUG_TOKEN, "POST-lp=%x\n", *lp);
	return (stmt);
}

/*
 * get_source_string - Expect a string variable (non-subscripted) or a string
 *			literal.
 */

uint8_t *
get_source_string (uint8_t *lp, uint8_t *varname, uint8_t *vartype, uint8_t *litstring, uint32_t *stringlen)
{
	uint8_t		*nlp;

	nlp = get_next_character (lp);

	*varname = 0;
	*vartype = 0;

	nlp = get_string_variable (lp, varname, vartype);
	if (lp != nlp) {
		lp = nlp;
		*stringlen = 9999;
		return (nlp);
	}
	/* Look for a string literal */
	if (*lp != '"' && *lp != '\'') {
		varname = 0;
		*stringlen = 9999;
		return (lp);
	}
	nlp = get_string_literal (lp, litstring, stringlen);
	if (lp == nlp) {
		varname = 0;
		return (lp);
	}
	Dprintf (DEBUG_TOKEN, "*** Got a string literal, stringlen=%d  litstring=\"%s\"\n", *stringlen, litstring);
	return (nlp);
}

/*
 * get_string_literal - Expect and store a string literal, in any combination
 *			of "XXX" or 'NNN.  litstring should be a pre-allocated
 *			uint8_t area for storage of the literal.  stringlen will
 *			contain the resultant length of characters.  Returns
 *			the new lp pointer, or the original lp pointer if no
 *			literal found.
 */

uint8_t *
get_string_literal (uint8_t *lp, uint8_t *litstring, uint32_t *stringlen)
{
	uint8_t		*nlp, *cnlp;
	uint32_t	cval;
	uint32_t	quoteopen = 0;
	uint32_t	stringfound = 0;
	uint32_t	only_singlequoted = 0;

	/* TSB separates leading single-quote string literals from any following double-quote literals in PRINT*/
	if (*stringlen == 0xdeadb00f)
		only_singlequoted = 1;

	*stringlen = 0;
	nlp = get_next_character (lp);
	while (*nlp) {
		if (*nlp == '\'' && !quoteopen) {
			nlp = get_next_character (nlp + 1);
#ifndef STANDARDLIT
			cnlp = get_number (nlp, GN_INT, &cval);
			if (cnlp == nlp || cval > 255 || cval < 0) {
				*stringlen = 9999;
				return (lp);
			}
			nlp = cnlp;
#else
			if (!isdigit(*nlp)) {
				*stringlen = 9999;
				return (lp);
			}
			cval = *nlp - '0';
			nlp = get_next_character (nlp + 1);
			if (isdigit(*nlp) && cval != 7) {
				/* Only check further digits if not already '7 (BELL) for 2000F->ACCESS bug. */
				cval = cval * 10 + (*nlp - '0');
				nlp = get_next_character (nlp + 1);
				if (isdigit(*nlp)) {
					if (isdigit(*get_next_character(nlp + 1))) {
						/*
					 	* We have a quoted literal > 3 digits.  Assume 2 digits.
					 	* This is a fixup for a bug in 2000F->ACCESS conversions.
					 	*/
						
					} else {
						cval = cval * 10 + (*nlp - '0');
						nlp = get_next_character (nlp);
					}
				}
			}
#endif
			litstring[*stringlen] = cval;
			*stringlen += 1;
			stringfound++;
		} else if (*nlp == '"' && !only_singlequoted) {
			if (quoteopen > 0) {
				quoteopen = 0;
				nlp = nlp + 1;
			} else {
				nlp = nlp + 1;
				quoteopen = 1;
				stringfound++;
				while (*nlp) {
					if (*nlp == '"') {
						nlp = nlp + 1;
						quoteopen = 0;
						break;
					}
					litstring[*stringlen] = *nlp;
					*stringlen += 1;
					nlp = nlp + 1;
				}
			}
		} else
			break;
	}
	if (stringfound)
		return (nlp);
	return (lp);
}

/*
 * get_relational_operator - Expect a relational operator (=, #, <>, >=, <=,
 *			      > or <) and set sttype to the operator code.
 *			      Returns the new lp pointer when found, or the
 *			      original lp if not found.
 */

uint8_t *
get_relational_operator (uint8_t *lp, uint32_t *sttype)
{
	uint8_t		*nlp, *nlp2;
	nlp = get_next_character (lp);
	if ((nlp2 = match_literal (nlp, "<>")) != NULL) {
		*sttype = 036;
		nlp = nlp2;
	} else if ((nlp2 = match_literal (nlp, ">=")) != NULL) {
		*sttype = 037;
		nlp = nlp2;
	} else if ((nlp2 = match_literal (nlp, "<=")) != NULL) {
		*sttype = 040;
		nlp = nlp2;
	} else if (*nlp == '>') {
		*sttype = 025;
		nlp = get_next_character (nlp + 1);
	} else if (*nlp == '<') {
		*sttype = 026;
		nlp = get_next_character (nlp + 1);
	} else if (*nlp == '#') {
		*sttype = 027;
		nlp = get_next_character (nlp + 1);
	} else if (*nlp == '=') {
		*sttype = 030;
		nlp = get_next_character (nlp + 1);
	} else
		return (lp);
	return (nlp);
}

#ifdef NOTUSED
/*
 * ans_source_string - Analyze And Store Source String.
 *			Expect and store either a string variable or literal
 *			and store it in stmt.  Returns a new stmt pointer
 *			or NULL if no valid match found.  String variable
 *			subscripts are accomodated.
 */

uint8_t *
ans_source_string (uint8_t *lp, uint8_t *stmt, uint32_t *stlen, uint32_t *sttype, uint8_t *varname, uint8_t *vartype)
{
	uint32_t	stringlen;
	uint8_t		litstring[512];
	uint8_t		*nlp;
	uint8_t		*nstmt;
	uint8_t		varop15;

	nlp = get_source_string (lp, varname, vartype, litstring, &stringlen);
	if (stringlen != 9999) {
		*stlen += 1;
		*stmt++ = ITOWH ((*sttype << 9));
		*stmt++ = ITOWL ((*sttype << 9));
		*sttype = 1;
		*stlen += 1;
		*stmt++ = ITOWH (*sttype << 9 | (stringlen & 0377));
		*stmt++ = ITOWL (*sttype << 9 | (stringlen & 0377));
		memcpy (stmt, litstring, stringlen);

		stmt += stringlen;
		*stlen += stringlen / 2;
		if (stringlen % 2) {
			*stmt++ = 0;
			*stlen += 1;
		}
		lp = nlp;
	} else if (nlp != lp) {
		lp = nlp;

		varop15 = 0;
		nstmt = get_subscript (&lp, sttype, stmt, stlen, varname, vartype, &varop15);
		if (nstmt == NULL)
			return (NULL);
		stmt = nstmt;
	}
	return (nlp);
}
#endif
