/*
 * tsb.h -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

uint32_t	tape_mach_type;
uint32_t	tape_system_level,
		tape_feature_level,
		tape_reelno,
		tape_date_year,
		tape_date_hour,
		tape_date_tsec;

#define TAPE_TYPE_UNKNOWN	0
#define TAPE_TYPE_SLEEP		1
#define TAPE_TYPE_HIBERNATE	2
#define TAPE_TYPE_DUMP		3

uint8_t		tape_type = 0;

uint8_t		extract = 0,
		create = 0,
		tokenize = 0,
		convert = 0,
		display = 0,
		verbose = 0,
		overwrite = 0,
		writedir = 0;
uint32_t	debug = 0;
uint8_t		only_programs = 0;
uint8_t		only_csave = 0;
uint8_t		only_files = 0;
uint8_t		only_ascii_files = 0;
uint8_t		all_entries = 1;
uint32_t	convtext = 0;

uint32_t	FNparse = 0;
uint8_t		FNvarname;
uint8_t		FNvartype;

uint32_t	forcelevel = 0;

uint8_t		IDtable[1024 * 80];
uint32_t	IDtablesize = 0;
uint8_t		gotIDrecord = 0;
uint32_t	Srecno;
uint32_t	Dtab_total;

uint32_t	program_entry_state;
char		*progname;
char		Nfilename[MAXPATHLEN] = "";
char		infilename[10][MAXPATHLEN];
int		fncount = -1;
int		fnno = 0;
char		buf[BUFMAX];
uint32_t	quoteflag = 0;

struct equiptable_2000access	eqAC;
struct equiptable_2000C		eqC;
struct equiptable_2000F		eqF;
struct equiptable_2000E		eqE;
struct equiptable_2000A		eqA;
struct DIRECtable_2000access	Dt2000access;
struct NONSHAREDEVICEtable_2000access NSDEV2000access;
struct SystemSegmentTable	SST;
struct SystemLibraryTable	SLT;
struct BlockTable {
	uint32_t		len;
	uint8_t *		block;
	struct BlockTable *	next;
} * DIRtab, * Dtab, * Dtabp, * pl;

uint8_t		* nDtab, * nDtabp;

struct Program_Entry Pentry;

#define ENTRY_TYPE_SET_PROT	8
#define ENTRY_TYPE_SET_ASCII	4
#define ENTRY_TYPE_SET_FILE	2
#define ENTRY_TYPE_SET_CSAVE	1
#define ENTRY_TYPE_SET_PROGRAM	0
#define ENTRY_TYPE_PROT(x)	(x & ENTRY_TYPE_SET_PROT)
#define ENTRY_TYPE_ASCII(x)	(x & ENTRY_TYPE_SET_ASCII)
#define ENTRY_TYPE_FILE(x)	(x & ENTRY_TYPE_SET_FILE)
#define ENTRY_TYPE_CSAVE(x)	(x & ENTRY_TYPE_SET_CSAVE)
#define ENTRY_TYPE_PROGRAM(x)	(!( x & ENTRY_TYPE_SET_ASCII) && !(x & ENTRY_TYPE_SET_FILE))

uint8_t *	Atrack[203][48];		/* 2000A Disk Image for program extraction */
						/* 203 Tracks x 48 sectors.  Allocate 256  */
						/* bytes to each sector when populating.   */

uint8_t *	Etrack[203][48];		/* 2000E Disk Image for program extraction */
						/* 203 Tracks x 48 sectors.  Allocate 256  */
						/* bytes to each sector when populating.   */

uint8_t *       feature_level_name[] = {
                (unsigned char *) "UNKNOWN",
                (unsigned char *) "200 - 2000/C (High-Speed Option)",
                (unsigned char *) "200 - 2000/F (All Options)",
                (unsigned char *) "1000 - HP 2000 Operating System, Release A (ACCESS)",
                (unsigned char *) "1500 - HP 2000 Operating System, Release B (ACCESS)",
                (unsigned char *) "2000 - HP 2000 Operating System, Release B (ACCESS)"
};

uint8_t *	file_flags[16] = {
        (unsigned char *) "Unrestricted",
        (unsigned char *) "Protected",
        (unsigned char *) "Locked",
        (unsigned char *) "Private",
        (unsigned char *) "[Undefined flag 4]",
        (unsigned char *) "[Undefined flag 5]",
        (unsigned char *) "[Undefined flag 6]",
        (unsigned char *) "[Undefined flag 7]",
        (unsigned char *) "[Undefined flag 8]",
        (unsigned char *) "[Undefined flag 9]",
        (unsigned char *) "[Undefined flag 10]",
        (unsigned char *) "FCP", 
        (unsigned char *) "MWA",
        (unsigned char *) "PFA",
        (unsigned char *) "Output",
        (unsigned char *) "Input"
};

uint8_t *        system_level_name[] = {
                (unsigned char *) "UNKNOWN",
                (unsigned char *) "1100 - HP-2000/A",
                (unsigned char *) "1200 - HP-2000/B",
                (unsigned char *) "1300 - HP-2000/C",
                (unsigned char *) "1500 - HP-2000/E",
                (unsigned char *) "2000 - HP-2000/C (High-Speed Option)",
		(unsigned char *) "2500 - HP-2000/C TSC-200 Extensions",
                (unsigned char *) "3000 - HP-2000/F (Options 200/205/500/505)",
                (unsigned char *) "3500 - HP-2000/F (Options 210/215/510/515) or (options 002/003)",
                (unsigned char *) "5000 - HP 2000/G Operating System, Release A (ACCESS)",
                (unsigned char *) "6000 - HP 2000 Operating System, Release B (ACCESS)"
};

struct debuglist {
	int	level;
	char	letter;
	char	*name;
} debuglist[] = {
	{ 3,	'H',	"headers"	},
	{ 4,	'S',	"system"	},
	{ 5,	'I',	"indconv"	},
	{ 6,	'L',	"tables"	},
	{ 7,	'R',	"fileread"	},
	{ 8,	'F',	"files"		},
	{ 9,	'M',	"memory"	},
	{ 10,	'P',	"tape"		},
	{ 11,	'D',	"detoken"	},
	{ 12,	'C',	"convert"	},
	{ 13,	'E',	"detoken2"	},
	{ 14,	'T',	"token"		},
	{ 15,	'A',	"ascii"		},
	{ 16,	'B',	"tapeblocks"	},
	{ 0,	0,	NULL		}
};
